# Copyright 2024 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Module with exceptions that may occur while working with Micetro."""
from requests import Response

from bluecat_libraries.http_client import ErrorResponse

__all__ = ["MicetroV2ErrorResponse"]


class MicetroV2ErrorResponse(ErrorResponse):
    """
    This exception is raised when an error response is received from a target.
    The HTTP response is available via the `response` instance member.
    """

    def __init__(self, message: str, response: Response, code: str, *args, **kwargs):
        super().__init__(message, response, *args, **kwargs)
        self.code = code

    def __repr__(self):
        return f"{self.__class__.__name__}({self.message!r} [{self.code}])"



# Copyright 2024 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Modules for working with Micetro."""



# Copyright 2024 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values for Client V2 in Micetro."""


class MediaType:
    """Values for the media type in header"""

    JSON = "application/json"
    XML = "application/xml"


class ResponseMessage:
    """Values for the response message returned by Micetro REST API"""

    INVALID_USERNAME_OR_PASSWORD = "Invalid username or password."
    MISSING_SESSION_ID = "Missing Session ID."



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Structures and functionality related to describing an HTTP client's target instance."""

import urllib3
from urllib3.util.url import Url
import logging
from typing import Optional


__all__ = [
    "Instance",
]

_DEFAULT_SCHEME = "https"


class Instance:
    """
    Definition of a service instance that can be used as a target by a client.

    :param url: URL of the service.
    :param logger: Logger to be used for logging. This value is currently not
        utilized.

    :ivar url: URL of the service.
    :ivar api_url_base: URL to be used as a base for constructing paths to the
        service's endpoints.

    .. versionadded:: 20.6.1
    """

    def __init__(self, url: str, logger: Optional[logging.Logger] = None):
        # pylint: disable=unused-private-member
        self.__logger = logger if logger else logging.getLogger(__name__)
        self.url: Url = urllib3.util.url.parse_url(url)
        self.url = self.url._replace(
            scheme=self.url.scheme if self.url.scheme else _DEFAULT_SCHEME,
            host=self.url.host if self.url.host else "localhost",
            query=None,
            fragment=None,
        )
        self.api_url_base: Url = self.url
        self.parse_url()

    def parse_url(self):
        """
        Process the service's URL and construct the value of the base URL of
        the service's API.
        """



# Copyright 2024 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Modules for working with Micetro's REST API."""
from bluecat_libraries.micetro.apiv2.client import Client
from bluecat_libraries.micetro.apiv2.constants import MediaType
from bluecat_libraries.micetro.apiv2.exceptions import MicetroV2ErrorResponse

__all__ = ["Client", "MediaType", "MicetroV2ErrorResponse"]



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Modules for working with BlueCat Address Manager's Failover API."""
from typing import Optional

from bluecat_libraries.http_client import Client as HTTPClient
from bluecat_libraries.address_manager.failover_api.instance import FailoverAPIInstance
from typing import Union


class Client(HTTPClient):
    """
    A client for calling the Failover API of Address Manager.

    The Failover API is a service configuration in Address Manager that can
    enable an externally automated failover when the primary Address Manager
    fails.

    :param url: The base URL of BAM's Failover service. For security reasons,
        the Failover Client always use HTTPS, even if the URL specifies an HTTP scheme.
        If no port is specified within the URL, the client will use port 8444.
    :param cert: Path to an SSL client certificate. The first item being the path to the certificate file
        and the second item is the key.
    :param verify: Whether to verify the server's TLS certificate. If a
        string is passed, it is treated as a path to a CA bundle to
        use. Defaults to ``True``.

    .. versionadded:: 21.5.1
    """

    def __init__(self, url: str, cert: tuple, verify: Union[bool, str] = True):
        super().__init__(target=FailoverAPIInstance(url=url))
        self.session.cert = cert
        self.session.headers = {"Accept": "application/json"}

        self.session.verify = verify

    def get_v1_health(self, timeout: Optional[int] = None) -> dict:
        """
        Return the health status of Address Manager.
        This API can be called on primary, secondary, or standalone Address Manager servers.

        :param timeout: The timeout of this request in milliseconds.
        :return: Returns the health status of Address Manager.

        Example:

            .. code-block:: python

                from bluecat_libraries.address_manager.failover_api import Client

                client_cert = ('client_certificate.crt', 'key.pem')
                server_cert = 'server_certificate.crt'
                with Client(url='https://<bam_address>:8444', cert=client_cert, verify=server_cert) as client:
                    data = client.get_v1_health()
                print(data)

        .. versionadded:: 21.5.1
        """
        params = None
        if timeout:
            params = {"timeout": timeout}
        response = self.http_get(
            url=self._url_for("/bam/v1/health"),
            params=params,
            cert=self.session.cert,
            verify=self.session.verify,
        )
        return response.json()

    def put_v1_promote(self) -> str:
        """
        Promote a secondary Address Manager server in the cluster to primary.

        :return: Returns the health status of Address Manager if successful.

        Example:

            .. code-block:: python

                from bluecat_libraries.address_manager.failover_api import Client

                client_cert = ('client_certificate.crt', 'key.pem')
                server_cert = 'server_certificate.crt'
                with Client(url='https://<bam_address>:8444', cert=client_cert, verify=server_cert) as client:
                    result = client.put_v1_promote()
                print(result)

        .. versionadded:: 21.5.1
        """
        response = self.http_put(
            url=self._url_for("/bam/v1/promote"), cert=self.session.cert, verify=self.session.verify
        )
        return response.json()

    def put_v1_managed_servers_takeover(self) -> list:
        """
        Notify all DNC/DHCP Servers controlled by the old primary Address Manager server
        of the new primary Address Manager server.

        :return: list

        Example:

            .. code-block:: python

                from bluecat_libraries.address_manager.failover_api import Client

                client_cert = ('client_certificate.crt', 'key.pem')
                server_cert = 'server_certificate.crt'
                with Client(url='https://<bam_address>:8444', cert=client_cert, verify=server_cert) as client:
                    data = client.put_v1_managed_servers_takeover()
                print(data)

        .. versionadded:: 21.5.1
        """
        response = self.http_put(
            url=self._url_for("/bam/v1/managed-servers/takeover"),
            cert=self.session.cert,
            verify=self.session.verify,
        )
        return response.json()

    def put_v1_managed_servers_takeover_by_id(self, server_id: int):
        """
        Notify a single DNC/DHCP Server controlled by the old primary Address Manager server
        of the new primary Address Manager server.

        :param server_id: The ID of the DNS/DHCP Server.
        :return: object

        Example:

            .. code-block:: python

                from bluecat_libraries.address_manager.failover_api import Client

                client_cert = ('client_certificate.crt', 'key.pem')
                server_cert = 'server_certificate.crt'
                server_id = 123
                with Client(url='https://<bam_address>:8444', cert=client_cert, verify=server_cert) as client:
                    data = client.put_v1_managed_servers_takeover_by_id(server_id)
                print(data)

        .. versionadded:: 21.5.1
        """
        response = self.http_put(
            url=self._url_for(f"/bam/v1/managed-servers/{server_id}/takeover"),
            cert=self.session.cert,
            verify=self.session.verify,
        )
        return response.json()



# Copyright 2023 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Modules for working with BlueCat Address Manager's API v2."""
from bluecat_libraries.address_manager.apiv2.client import Client
from bluecat_libraries.address_manager.apiv2.constants import MediaType
from bluecat_libraries.address_manager.apiv2.exceptions import BAMV2ErrorResponse

__all__ = ["BAMV2ErrorResponse", "Client", "MediaType"]



# Copyright 2020-2023 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# pylint: disable=C0112,C0115,C0116,W0511,W0613;
"""Functions related to the conversion of data to and from BlueCat Address Manager's API."""
import re
from typing import Union

from bluecat_libraries.http_client.exceptions import GeneralError


def deserialize_joined_key_value_pairs(value: str, item_sep="|", keyvalue_sep="=") -> dict:
    """Convert a single-string representation into a dictionary."""
    # NOTE: A straightforward "split by pipes and then split by equal signs" would cause erroneous interpretations,
    # because of the presence of escaped pipes in both names and values.
    # NOTE: Currently, equal signs are not escaped, although they may be present in both the name and the value of a
    # property. We can have "=====" as a key-value pair and no way of knowing is it "=" : "===", "==": "==", or
    # "===": "=". Thus, the first occurring equal sign is treated as a separator.
    properties = {}
    data = []
    is_escape = False
    key_name = None
    for char in value:
        if char == keyvalue_sep and not key_name:
            key_name = "".join(data)
            data = []
        elif char == item_sep and not is_escape:
            if key_name:
                properties[key_name] = "".join(data)
                key_name = None
            data = []
        elif char == "\\" and not is_escape:
            is_escape = True
            continue
        else:
            if is_escape and char != item_sep:
                data.append("\\")
            data.append(char)
        is_escape = False
    if key_name:
        properties[key_name] = "".join(data)
    return properties


def serialize_joined_key_value_pairs(values: dict, item_sep="|", keyvalue_sep="=") -> str:
    if not values:
        return ""

    result = item_sep.join(
        "{key}{keyvalue_sep}{value}".format(
            key=escape(check_allowed_condition(key)),
            keyvalue_sep=keyvalue_sep,
            value=escape(check_allowed_condition(str(value))),
        )
        for key, value in values.items()
    )
    return result + item_sep if result else ""


def serialize_joined_values(values: list, item_sep=",") -> str:
    if not values:
        return ""
    result = item_sep.join(str(x) for x in values)
    return result


def serialize_possible_list(value: Union[str, list[str], list[list[str]]]) -> str:
    """
    If ``value`` is a list, then either all inner elements must be lists, or no element can be a
    list. Otherwise, no behavior is guaranteed.
    """
    # Simple value case
    if not isinstance(value, list):
        return str(value)

    # Empty list case
    if not value:
        return ""

    # 1D case
    if not isinstance(value[0], list):
        return ",".join(str(x) for x in value)

    # 2D case
    out = []
    for inner in value:
        x = ",".join(str(x) for x in inner)
        out.append("{" + x + "}")
    return ",".join(out)


def deserialize_possible_list(value: str) -> Union[str, list[str], list[list[str]]]:
    # We rely on the fact that individual values do not contain `{`, `}`, or `,`

    # 2D case
    if value.startswith("{"):
        return [item.group(1).split(",") for item in re.finditer(r"{([^}]*)}", value)]

    # Empty string becomes [""]
    out = value.split(",")
    return out if len(out) > 1 else out[0]


def escape(value: str) -> str:
    return value.replace("|", "\\|")


def unescape(value: str) -> str:
    return value.replace("\\|", "|")


def check_allowed_condition(value: str) -> str:
    """
    Check whether the value is acceptable.

    :param value: The value to be checked.
    :raises GeneralError: If the value is not acceptable.
    """
    # NOTE: BAM API v1 uses the pipe character - "|" - as separator between
    # key-value pairs. Additionally, when the pipe character has to be present
    # in a value it is "escaped" by preceding it with a backslash - "\".
    # Because backslashes themselves are not escaped, if a value ends up with
    # one, then the API will consider the pipe as escaped. That would result in
    # the whole next key-value pair to be considered as part of the value of the
    # current pair.
    # Thus, to avoid creating values in BAM different from what the user
    # intended, we do not accept values to end with a backslash.
    if value.endswith("\\"):
        raise GeneralError("Backslash is not allowed at the end of a value.")
    return value



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in POST /v1/addServer method in BlueCat Address Manager."""
from ._enum import StrEnum


class ServerCapabilityProfiles(StrEnum):
    """
    Values for types of server profiles in BlueCat Address Manager.

    Following changes to supported server profiles in Integrity 9.6.0,
    some profiles in this class are not supported as of Integrity 9.6.0+.
    Others were introduced, and are only supported with Integrity 9.6.0+.

    Below is a table that lists which profiles are compatible with which Integrity
    versions

    .. list-table::
        :header-rows: 1

        * - Profile name
          - Supported prior to 9.6.0
          - Supported by 9.6.0 and later
        * - ADONIS_800
          - Yes
          - No
        * - ADONIS_1200
          - Yes
          - No
        * - ADONIS_1900
          - Yes
          - No
        * - ADONIS_1950
          - Yes
          - No
        * - ADONIS_1950
          - Yes
          - No
        * - ADONIS_XMB2
          - Yes
          - No
        * - ADONIS_XMB3
          - Yes
          - Yes
        * - AFILIAS_DNS_SERVER
          - Yes
          - Yes
        * - BLUECAT_ENTERPRISE
          - No
          - Yes
        * - BLUECAT_CORE
          - No
          - Yes
        * - BLUECAT_BRANCH_H
          - No
          - Yes
        * - BLUECAT_BRANCH_L
          - No
          - Yes
        * - DNS_DHCP_SERVER_20
          - Yes
          - Yes
        * - DNS_DHCP_SERVER_45
          - Yes
          - Yes
        * - DNS_DHCP_SERVER_60
          - Yes
          - Yes
        * - DNS_DHCP_SERVER_100
          - Yes
          - Yes
        * - DNS_DHCP_SERVER_100_D
          - Yes
          - Yes
        * - DNS_DHCP_GEN4_2000
          - Yes
          - Yes
        * - DNS_DHCP_GEN4_4000
          - Yes
          - Yes
        * - DNS_DHCP_GEN4_5000
          - Yes
          - Yes
        * - DNS_DHCP_GEN4_7000
          - Yes
          - Yes
        * - LEGACY
          - No
          - Yes
        * - OTHER_DNS_SERVER
          - Yes
          - Yes

    All profiles that were removed with Integrity 9.6.0+ are replaced with
    the ``LEGACY`` profile, which represents any device that now doesn't have a supported
    profile.
    """

    ADONIS_800 = "ADONIS_800"  # replaced with "LEGACY" as of Integrity 9.6.0
    ADONIS_1200 = "ADONIS_1200"  # replaced with "LEGACY" as of Integrity 9.6.0
    ADONIS_1900 = "ADONIS_1900"  # replaced with "LEGACY" as of Integrity 9.6.0
    ADONIS_1950 = "ADONIS_1950"  # replaced with "LEGACY" as of Integrity 9.6.0
    ADONIS_XMB2 = "ADONIS_XMB2"  # replaced with "LEGACY" as of Integrity 9.6.0
    ADONIS_XMB3 = "ADONIS_XMB3"
    AFILIAS_DNS_SERVER = "AFILIAS_DNS_SERVER"
    BLUECAT_ENTERPRISE = "BLUECAT_ENTERPRISE"  # new in Integrity 9.6.0
    BLUECAT_CORE = "BLUECAT_CORE"  # new in Integrity 9.6.0
    BLUECAT_BRANCH_H = "BLUECAT_BRANCH_H"  # new in Integrity 9.6.0
    BLUECAT_BRANCH_L = "BLUECAT_BRANCH_L"  # new in Integrity 9.6.0
    DNS_DHCP_SERVER_20 = "DNS_DHCP_SERVER_20"
    DNS_DHCP_SERVER_45 = "DNS_DHCP_SERVER_45"
    DNS_DHCP_SERVER_60 = "DNS_DHCP_SERVER_60"
    DNS_DHCP_SERVER_100 = "DNS_DHCP_SERVER_100"
    DNS_DHCP_SERVER_100_D = "DNS_DHCP_SERVER_100_D"
    DNS_DHCP_GEN4_2000 = "DNS_DHCP_GEN4_2000"
    DNS_DHCP_GEN4_4000 = "DNS_DHCP_GEN4_4000"
    DNS_DHCP_GEN4_5000 = "DNS_DHCP_GEN4_5000"
    DNS_DHCP_GEN4_7000 = "DNS_DHCP_GEN4_7000"
    LEGACY = "LEGACY"  # new in Integrity 9.6.0, catchall for legacy devices
    OTHER_DNS_SERVER = "OTHER_DNS_SERVER"



# Copyright 2023 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values for Client V2 in Bluecat libraries."""


class MediaType:
    """Values for the media type in header"""

    CSV = "text/csv"
    FORM_DATA = "multipart/form-data"
    HAL_JSON = "application/hal+json"
    HTML = "text/html"
    JSON = "application/json"
    MERGE_PATCH_JSON = "application/merge-patch+json"
    OCTET_STREAM = "application/octet-stream"



# Copyright 2024 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Module for the client to work with Micetro's REST API."""
import base64
import requests

from json import JSONDecodeError
from typing import Any, Callable, Iterable, IO, Optional, Union
from urllib3.util import parse_url

from bluecat_libraries.http_client import (
    Client as HTTPClient,
    ClientError,
    Instance,
    GeneralError,
    UnexpectedResponse,
    version,
)
from bluecat_libraries.micetro.apiv2.constants import ResponseMessage, MediaType
from bluecat_libraries.micetro.apiv2.exceptions import MicetroV2ErrorResponse


__all__ = ["Client"]


class _Instance(Instance):
    """
    Definition of the Micetro REST API service. It is used as a target by a client.

    It holds a URL to the root of the Micetro.
    It holds a URL to the root of the Micetro REST API.

    The path, query, and fragments portions of the input URL will be ignored,
    even if provided.

    .. note:: For internal use *only*.

    :meta private:

    .. versionadded:: 24.3.0
    """

    def parse_url(self):
        """
        Process the service's URL and construct the value of the base URL of
        the service's API.
        """
        self.url = self.url._replace(auth=None, path=None, query=None, fragment=None)
        self.api_url_base = self.url._replace(path="/mmws/api/v2")


class Client(HTTPClient):
    """
     A client for making HTTP calls to Micetro REST API.

     There is a method for each HTTP verb - GET, POST, PUT, PATCH, and DELETE -
     as well as a generic method for a HTTP request: http_get, http_post,
     http_put, http_patch, http_delete, and http_request. The functionality of
     the client relies on the use of these methods.

     The benefits of this client over a direct use of the ``requests`` Python
     library are:

         1.  Available methods for performing authentication.
         2.  The client keeps track of the authentication token and automatically
             sends it with each request.
         3.  The client keeps track of the base API URL, needing only the
             relative path to be specified when making a request.
         4.  The client automatically parses the responses and returns Python
             data structures - dictionaries or strings - based on the response
             content type.
         5.  The client detects error responses and raises a Python exception,
             which holds the fields that the Micetro API replied with, e.g, code.

     Overall, users of the client can write less code and keep track of less data
     when making requests to the Micetro REST API.

     Nonetheless, the client allows for taking advantage of the full
     functionality of the Micetro REST API by exposing all available ``requests``
     parameters in its methods. That way a user can, for example, specify filters
     and fields selection through URL query parameters.

    You need to authenticate to the Micetro REST API via the client method (with a username
     and a password) when making calls. The client will use HTTP header ``Authorization`` as per the
     Micetro documentation.

     .. note:: The client does not specify a value for the `Accept` header by
         default. This results in the `Content-Type` of the response to be
         determined by the defaults of Micetro.

     Example:

     .. code-block:: python

         from bluecat_libraries.micetro.apiv2 import Client, MediaType

         # Retrieve the users. Request the data as per Micetro's default content type.
         with Client(<micetro_url>) as client:
             client.authenticate(<username>, <password>)
             response = client.http_get("/users")
             users = response["result"]["users"]
             for user in users:
                 print(f'{user["ref"]}: {user["name"]}')

         # Retrieve the users. Request that the response is in 'XML' format.
         with Client(<micetro_url>) as client:
             client.authenticate(<username>, <password>)
             response = client.http_get(
                 "/users",
                 headers={"Accept": MediaType.XML},
             )
             print(response)

     .. versionadded:: 24.3.0
    """

    def __init__(self, url: str, *, verify: Any = True) -> None:
        u = parse_url(url)
        if not u.host:
            raise ClientError("The Micetro URL does not contain a host. It is required.")

        super().__init__(_Instance(url))
        self.session.verify = verify
        self.__version = None  # The version of Micetro

    # region Authentication

    @staticmethod
    def generate_token(username, password) -> str:
        """
        Generate a token following the Basic Authentication standard.

        :param username: The username for the Micetro account.
        :param password: The password associated with the username.
        """
        credentials = f"{username}:{password}"
        credentials = base64.b64encode(credentials.encode("utf-8")).decode("utf-8")
        return f"Basic {credentials}"

    def clear_token(self) -> None:
        """Clear token from client and session header"""
        self.token_type = None
        self.token = None
        if "Authorization" in self.session.headers:
            del self.session.headers["Authorization"]

    def authenticate(self, username: str, password: str, refresh_token: bool = False) -> str:
        """
        Logs into Micetro using encoded credentials.

        Creates a base64-encoded token from the provided username and password,
        verifies it with Micetro to ensure the credentials are valid,
        and stores the token for use in subsequent API calls.

        When the client is already authenticated, this function behaves as follows:
            - If ``refresh_token`` is ``False``, it raises a ``ClientError('Client is already authenticated.')``.
            - If ``refresh_token`` is ``True``, it returns a newly generated token.

        :param username: The username for the Micetro account.
        :param password: The password associated with the username.
        :param refresh_token: Retrieve a new token even if logged in.
        :return: A string that contains the authentication token.

        Example:

            .. code:: python

                from bluecat_libraries.micetro.apiv2 import Client

                with Client(<micetro_url>) as client:
                    client.authenticate(<micetro_username>, <micetro_password>)

        .. versionadded:: 24.3.0
        """
        if self.is_authenticated and not refresh_token:
            raise ClientError("Client is already authenticated.")

        self.auth = self.generate_token(username, password)
        return self.auth

    # endregion Authentication

    # region Inherited interfaces

    def _url_for(self, name: str) -> str:
        # NOTE: The `self.api_url_base` does not end with `/`.
        # The value of `name` is expected to start with `/`.
        return self.target.api_url_base.url + name

    def _handle_request_exc(self, exc: Exception) -> None:
        """
        Handle the exception that is raised when the request is attempted.

        :param exc: The exception raised when the HTTP request is made.
        :raises GeneralError: When the request to Micetro fails.

        This method is not intended to be called by users of the class. It is
        present to modify the behaviour inherited from the base class.
        """
        raise GeneralError("The request to Micetro REST API v2 failed.") from exc

    def _handle_error_response(self, response: requests.Response) -> None:
        """
        Handle a response that's considered to be an error (based on the HTTP
        status code). The standard implementation raises an exception.

        :param response: Object representing the HTTP response.

        .. versionchanged:: 24.3.0
        """
        response_content_type = response.headers.get("Content-Type")
        if MediaType.JSON in response.headers["Content-Type"].lower():
            try:
                data = response.json()
            except Exception as exc:
                raise UnexpectedResponse(
                    "The Micetro REST API v2 error response is not valid JSON,"
                    " despite the specified content type.",
                    response,
                ) from exc
            raise MicetroV2ErrorResponse(
                message=data["error"]["message"],
                response=response,
                code=data["error"]["code"],
            )
        raise UnexpectedResponse(
            "The Micetro REST API v2 error response does not have JSON content"
            f" type as expected: {response_content_type}",
            response,
        )

    def _handle_nonerror_response(self, response: requests.Response) -> Union[str, dict]:
        """
        Handle a response that's considered to be a success (non error), based
        on the HTTP status code. Depending on the content type of the response,
        its body is either parsed as JSON and the resulting object returned, or
        returned as verbatim text.

        :param response: Object representing the HTTP response.
        :return: The result of the handled response.
        """
        response_content_type = response.headers.get("Content-Type")
        if response_content_type and MediaType.JSON in response_content_type:
            try:
                # At times, the responses from the PUT, POST, and DELETE methods in Micetro
                # may not include JSON data.
                content = response.json()
            except JSONDecodeError:
                content = response.text
            return content
        return response.text

    def http_request(
        self,
        method: str,
        url: str,
        params: Optional[Union[dict, list[tuple], bytes]] = None,
        data: Optional[Union[dict, list[tuple], bytes, IO]] = None,
        headers: Optional[dict] = None,
        cookies: Optional[dict] = None,
        files: Optional[dict[str, IO]] = None,
        auth: Optional[Union[tuple, Callable]] = None,
        timeout: Optional[Union[float, tuple]] = None,
        allow_redirects: bool = True,
        proxies: Optional[dict] = None,
        hooks: Optional[dict] = None,
        stream: Optional[bool] = None,
        verify: Union[bool, str] = None,
        cert: Optional[Union[str, tuple[str, str]]] = None,
        json: Optional[Any] = None,
        expected_status_codes: Optional[Iterable[int]] = None,
    ) -> Union[dict, str]:
        """
        Perform an HTTP request to the Micetro using the provided
        parameters.

        The call will be made through the instance of ``requests.Session`` that
        the client maintains. The parameters have the same behaviour as their
        namesakes in the ``requests`` library. The only exception is ``url``,
        whose value should be relative to the Micetro URL, e.g.,
        ``"/users"``.

        :param method: HTTP method to be used, e.g., ``GET``, ``POST``,
            ``PUT``, ``PATCH``, ``DELETE``.
        :param url: A URL, relative to the API root to be used for the request,
            e.g., ``/users``. It must start with a forward slash.
        :param params: Query parameters to be passed in the request.
        :param data: Value to send in the body of the request. It can be
            a dictionary, list of tuples, bytes, or file-like object.
        :param headers: HTTP headers to send with the request.
        :param cookies: Cookies to send with the request.
        :param files: File-like objects for multipart encoding upload, e.g.
            ``{"filename": file_like_object}``
        :param auth: Object to handle HTTP Authentication.
        :param timeout: How long to wait to receive data before giving up. If
            given as a tuple, it is used as (connect, read) timeouts.
        :param allow_redirects: Whether to allow redirects. Defaults to
            ``True``.
        :param proxies: Mapping of protocol or protocol and hostname to a URL
            of a proxy to be used.
        :param hooks: Hooks to be called upon events.
        :param stream: Whether to immediately read the response content.
            Defaults to ``False``.
        :param verify: Whether to verify the server's TLS certificate. If a
            string is passed, it is treated as a path to a CA bundle to be
            used. Defaults to ``True``.
        :param cert: Path to a SSL client certificate file (.pem). If the
            certificate has a key, it can be provided as the second item in a
            tuple, with the first item being the path to the certificate file.
        :param json: Object to be sent as JSON in the body of the request.
        :param expected_status_codes: HTTP status codes that are acceptable as
            a status code of the HTTP response. If the received code does not
            match the passed values, an :py:exc:`UnexpectedResponse` is raised.
            If left empty, the status code validation is not performed.
        :return: The data the API responded with. The client will process the
            response based on the stated content type and return either the
            JSON formatted data parsed into a Python object (e.g., a ``dict``)
            or the verbatim body text.

        Example:

        .. code-block:: python

            from bluecat_libraries.micetro.apiv2 import Client

            # Retrieve the users. Request the data as per Micetro's default content type.
            with Client(<micetro_host_url>) as client:
                client.login(<username>, <password>)
                response = client.http_request("GET", "/users")
                users = response["result"]["users"]
                for user in users:
                    print(f'{user["ref"]}: {user["name"]}')
        """
        url = self._url_for(url)
        return super().http_request(
            method,
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            expected_status_codes=expected_status_codes,
        )

    def http_get(
        self,
        url: str = "",
        params=None,
        expected_status_codes: Optional[Iterable[int]] = None,
        **kwargs,
    ) -> Union[dict, str]:
        """
        Perform an HTTP ``GET`` request based on the provided parameters.
        See method ``http_request`` for details.

        If ``expected_status_codes`` is not specified, it defaults to ``(200, )``.
        ``200`` is the HTTP status code for successful response ``OK``.
        """
        if expected_status_codes is None:
            expected_status_codes = (requests.codes.ok,)
        return self.http_request(
            "GET", url, params, expected_status_codes=expected_status_codes, **kwargs
        )

    def http_post(
        self,
        url: str,
        params=None,
        data=None,
        json: Optional[Any] = None,
        expected_status_codes: Optional[Iterable[int]] = None,
        **kwargs,
    ) -> Union[dict, str]:
        """
        Perform an HTTP ``POST`` request based on the provided parameters.
        See method ``http_request`` for details.
        """
        return self.http_request(
            "POST",
            url,
            params,
            data,
            json=json,
            expected_status_codes=expected_status_codes,
            **kwargs,
        )

    def http_put(
        self,
        url: str,
        params=None,
        data=None,
        json: Optional[Any] = None,
        expected_status_codes: Optional[Iterable[int]] = None,
        **kwargs,
    ) -> Union[dict, str]:
        """
        Perform an HTTP ``PUT`` request based on the provided parameters.
        See method ``http_request`` for details.
        """
        return self.http_request(
            "PUT",
            url,
            params,
            data,
            json=json,
            expected_status_codes=expected_status_codes,
            **kwargs,
        )

    def http_patch(
        self,
        url: str,
        params=None,
        data=None,
        json: Optional[Any] = None,
        expected_status_codes: Optional[Iterable[int]] = None,
        **kwargs,
    ) -> Union[dict, str]:
        """
        Perform an HTTP ``PATCH`` request based on the provided parameters.
        See method ``http_request`` for details.
        """
        return self.http_request(
            "PATCH",
            url,
            params,
            data,
            json=json,
            expected_status_codes=expected_status_codes,
            **kwargs,
        )

    def http_delete(
        self,
        url: str,
        params=None,
        data=None,
        json: Optional[Any] = None,
        expected_status_codes: Optional[Iterable[int]] = None,
        **kwargs,
    ) -> Union[dict, str]:
        """
        Perform an HTTP ``DELETE`` request based on the provided parameters.
        See method ``http_request`` for details.
        """
        return self.http_request(
            "DELETE",
            url,
            params,
            data,
            json=json,
            expected_status_codes=expected_status_codes,
            **kwargs,
        )

    # endregion Inherited interfaces

    # region Micetro REST v2 API specific interfaces

    @property
    def auth(self) -> Union[str, None]:
        """
        The value that the client is configured to use for header ``Authorization`` when making
        requests to REST API.
        """
        return self.session.headers.get("Authorization")

    @auth.setter
    def auth(self, value: str) -> None:
        """
        Set the value to use for header ``Authorization``. It is used for authenticating the
        requests to the Micetro REST API.

        :param value: Token to be used for Micetro authentication.
        """
        if not value:
            raise ClientError("Empty value for authentication to Micetro REST v2 API.")
        if not value.startswith("Basic"):
            raise ClientError("Micetro REST v2 API only supports Basic authentication.")
        try:
            self.http_get("/addressSpaces", headers={"Authorization": value}, params={"limit": 1})
        except MicetroV2ErrorResponse as exc:
            # These two exceptions explicitly describe the token as not acceptable by Micetro
            # RESTv2. Other exceptions indicating the user does not have required permission
            # will not raise during this authentication, they should be raised in the user's
            # next request and handled by http_request.
            if ResponseMessage.INVALID_USERNAME_OR_PASSWORD == str(
                exc
            ) or ResponseMessage.MISSING_SESSION_ID == str(exc):
                raise exc
        self.token_type = "Basic"
        self.token = value
        self.session.headers.update({"Authorization": value})

    @property
    def micetro_url(self) -> str:
        """
        A URL to the Micetro server the client is created for.

        This is not the verbatim value given during initialization. It contains
        only the scheme, host, and (if specified) port, which will be used by
        the client.
        """
        return self.target.url.url

    @property
    def micetro_version(self) -> version.Version:
        """
        The version of Micetro the client is connected to.

        .. note:: The client does not have to be authenticated to get version.
        """
        if self.__version is None:
            # This API is an alternative to the one `GET /micetro` which requires
            # administrative access to obtain version.
            # The response of this API depends on whether user authenticated or not,
            # but the `centralVersion` is always included. To avoid the inconsistency
            # and ensure authenticated users without any roles do not encounter exceptions
            # while retrieving version, thereby the `Authorization` in header is always
            # overridden, regardless of whether user is authenticated or not.
            response = self.http_get("/command/web_getPreLoginInfo", headers={"Authorization": ""})
            central_version = response["result"]["centralVersion"]
            self.__version = version.parse_version(central_version)
        return self.__version

    # endregion Micetro REST v2 API specific interfaces



# Copyright 2023-2024 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""
Utilities related to the version of system.

.. versionadded:: 23.1.0
"""
from __future__ import annotations
import collections


__all__ = ["Version", "parse_version"]


class Version:
    """
    A representation of the version of a system.

    :param major: The major part of the version. Defaults to ``0``.
    :param minor: The minor part of the version. Defaults to ``0``.
    :param patch: The patch part of the version. Defaults to ``0``.

    It consists of three integer values for the ``major``, ``minor``, and ``patch`` parts of the
    version. The parts can be addressed through their name or index. It supports comparison to
    strings, sequences, mappings.

    Example:

    .. code-block:: python

        >>> v1 = Version(11, 22, 33)
        >>> assert v1.major == 11
        >>> assert v1[2] == 33
        >>> assert str(v1) == "11.22.33"
        >>> assert repr(v1) == "Version(major=11, minor=22, patch=33)"
        >>> v2 = Version(minor=44)
        >>> assert v2 == (0, 44)
        >>> assert [0, 44, 0] == v2
        >>> assert v2 == {"patch": 0, "minor": 44}

    .. note::

        Slices of the version are tuples. They are different from slices of the string
        representation of the same version.

        .. code-block:: python

            >>> v1 = Version(11, 22, 33)
            >>> assert v1[:2] == (11, 22)
            >>> assert str(v1)[:2] == "11"

    .. versionadded:: 23.1.0
    """

    def __init__(self, major: int = 0, minor: int = 0, patch: int = 0):
        if not isinstance(major, int):
            raise TypeError(f"Argument 'major' must be an 'int', not '{type(major).__name__}'.")
        if not isinstance(minor, int):
            raise TypeError(f"Argument 'minor' must be an 'int', not '{type(minor).__name__}'.")
        if not isinstance(patch, int):
            raise TypeError(f"Argument 'patch' must be an 'int', not '{type(patch).__name__}'.")
        if major < 0:
            raise ValueError("Negative value for the 'major' part of the version.")
        if minor < 0:
            raise ValueError("Negative value for the 'minor' part of the version.")
        if patch < 0:
            raise ValueError("Negative value for the 'patch' part of the version.")
        self.__value = (major, minor, patch)

    @property
    def major(self) -> int:
        """The major part of the version."""
        return self.__value[0]

    @property
    def minor(self) -> int:
        """The minor part of the version."""
        return self.__value[1]

    @property
    def patch(self) -> int:
        """The patch part of the version."""
        return self.__value[2]

    def __str__(self) -> str:
        return ".".join([str(o) for o in self.__value])

    def __repr__(self) -> str:
        return "Version(major={}, minor={}, patch={})".format(*self.__value)

    def __getitem__(self, item):
        return self.__value[item]

    def __iter__(self):
        return iter(self.__value)

    def __cmp_helper(self, op: str, other: Version | str | tuple | list | dict) -> bool:
        """Generic comparison performing helper."""
        if isinstance(other, Version):
            o = other
        elif isinstance(other, str):
            o = parse_version(other)
        elif isinstance(other, collections.abc.Sequence):
            o = Version(*other)
        elif isinstance(other, collections.abc.Mapping):
            o = Version(**other)
        else:
            raise TypeError(
                f"Cannot compare a Version object to instances of '{type(other).__name__}'."
            )
        me = tuple(self)
        return getattr(me, op)(tuple(o))

    def __eq__(self, other: Version | str | tuple | list | dict) -> bool:
        return self.__cmp_helper("__eq__", other)

    def __ne__(self, other: Version | str | tuple | list | dict) -> bool:
        return self.__cmp_helper("__ne__", other)

    def __lt__(self, other: Version | str | tuple | list | dict) -> bool:
        return self.__cmp_helper("__lt__", other)

    def __le__(self, other: Version | str | tuple | list | dict) -> bool:
        return self.__cmp_helper("__le__", other)

    def __gt__(self, other: Version | str | tuple | list | dict) -> bool:
        return self.__cmp_helper("__gt__", other)

    def __ge__(self, other: Version | str | tuple | list | dict) -> bool:
        return self.__cmp_helper("__ge__", other)


def parse_version(value: str) -> Version:
    """
    Create a version object by parsing a string value.

    :param value: A version number.

        *   The version may have up to three parts.
        *   The version parts should be separated by periods (``.``).
        *   The version parts should be integers.

    :return: The parsed version.

    Example:

    .. code-block:: python

        >>> s = "11.22"
        >>> v = parse_version(s)
        >>> assert str(v) == "11.22.0"
        >>> assert repr(v) == "Version(major=11, minor=22, patch=0)"
        >>> assert v.major == 11
        >>> assert v[2] == 0

    .. versionadded:: 23.1.0
    """
    values = (int(o) for o in value.split("."))
    return Version(*values)



# Copyright 2023-2024 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""
Utilities related to the version of BlueCat Address Manager.

.. versionadded:: 23.1.0
"""
import warnings
from bluecat_libraries.http_client.version import Version, parse_version

# Warn about the deprecation
warnings.warn(
    "This module is moved to `http_client` and might be removed in future versions. "
    "Please update your imports.",
    ImportWarning,
)

__all__ = ["Version", "parse_version"]



# Copyright 2023 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Module with exceptions that may occur while working with BlueCat Address Manager."""
from requests import Response
from bluecat_libraries.http_client import ErrorResponse


__all__ = ["BAMV2ErrorResponse"]


class BAMV2ErrorResponse(ErrorResponse):
    """
     This exception is raised when an error response is received from a target.
     The HTTP response is available via the `response` instance member.

    :param message: Description of API error code.
    :param response: The HTTP response resulting from a request.
    :param status: HTTP status code of error (400-599).
    :param reason: Reason phrase of status code.
    :param code: API error code.

    Example:

    .. code-block:: python

        from bluecat_libraries.address_manager.apiv2 import Client, BAMV2ErrorResponse

        with Client(<bam_host_url>) as client:
            try:
                # This will fail because the authentication is missing.
                response = client.http_get("/configurations")
            except BAMV2ErrorResponse as exc:
                print(exc.message)
                print(exc.status)
                print(exc.reason)
                print(exc.code)

        >>> The request authorization credentials are either missing or invalid
        >>> 401
        >>> Unauthorized
        >>> InvalidAuthorizationCredentials
    """

    def __init__(
        self,
        message: str,
        response: Response,
        status: int,
        reason: str,
        code: str,
    ):
        super().__init__(message, response)
        self.status = status
        self.reason = reason
        self.code = code

    def __repr__(self):
        return f"{self.__class__.__name__}({self.message!r} [{self.code}, {self.reason}, {self.status}])"



# Copyright 2023-2024 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Module for the client to work with BlueCat Address Manager's API v2."""
from __future__ import annotations
from bluecat_libraries.address_manager.apiv2.constants import MediaType
from bluecat_libraries.address_manager.apiv2.exceptions import BAMV2ErrorResponse
from bluecat_libraries.http_client import (
    Client as HTTPClient,
    ClientError,
    GeneralError,
    Instance,
    UnexpectedResponse,
    version,
)
import copy
import requests
from urllib3.util import parse_url
from typing import Any, Callable, Iterable, IO, Optional, Union


__all__ = ["Client"]


class _Instance(Instance):
    """
    Definition of the BAM REST v2 API service. It is used as a target by a client.

    It holds a URL to the root of the BlueCat Address Manager.
    It holds a URL to the root of the BlueCat Address Manager REST API v2.

    The path, query, and fragments portions of the input URL will be ignored,
    even if provided.

    .. note:: For internal use *only*.

    :meta private:

    .. versionadded:: 23.1.0
    """

    def parse_url(self):
        """
        Process the service's URL and construct the value of the base URL of
        the service's API.
        """
        self.url = self.url._replace(auth=None, path=None, query=None, fragment=None)
        self.api_url_base = self.url._replace(path="/api/v2")


class Client(HTTPClient):
    """
    A client for making HTTP calls to BAM REST v2 API.

    There is a method for each HTTP verb - GET, POST, PUT, PATCH, and DELETE -
    as well as a generic method for a HTTP request: http_get, http_post,
    http_put, http_patch, http_delete, and http_request. The functionality of
    the client relies on the use of these methods.

    The benefits of this client over a direct use of the ``requests`` Python
    library are:

        1.  Available methods for performing login and logout, as well as
            determining the version of the BlueCat Address Manager.
        2.  The client keeps track of the authentication token and automatically
            sends it with each request.
        3.  The client keeps track of the base API URL, needing only the
            relative path to be specified when making a request.
        4.  The client automatically parses the responses and returns Python
            data structures - dictionaries or strings - based on the response
            content type.
        5.  The client detects error responses and raises a Python exception,
            which holds the fields that the BAM API replied with, e.g, code.

    Overall, users of the client can write less code and keep track of less data
    when making requests to the BAM REST v2 API.

    Nonetheless, the client allows for taking advantage of the full
    functionality of the BAM REST v2 API by exposing all available ``requests``
    parameters in its methods. That way a user can, for example, specify filters
    and fields selection through URL query parameters.

    You need to authenticate to the BAM REST v2 API when making calls. The
    client will use HTTP header ``Authorization`` as per the BAM documentation.
    You have to either perform a login via the client method (with a username
    and a password) or set the ``auth`` property to the full value for the
    header, including the authorization scheme.

    .. note:: The client does not specify a value for the `Accept` header by
        default. This results in the `Content-Type` of the response to be
        determined by the defaults of BAM.

    Example:

    .. code-block:: python

        from bluecat_libraries.address_manager.apiv2 import Client, MediaType
        import csv

        # Retrieve the configurations. Request the data as per BAM's default content type.
        with Client(<bam_host_url>) as client:
            client.login(<username>, <password>)
            response = client.http_get("/configurations")
            configurations = response["data"]
            for configuration in configurations:
                print(f'{configuration["id"]}: {configuration["name"]}')
            client.logout()

        # Retrieve the configurations. Request that the response is in 'JSON' format.
        # The result should contain only fields 'id' and 'name'.
        with Client(<bam_host_url>) as client:
            client.login(<username>, <password>)
            response = client.http_get(
                "/configurations",
                params={"fields": "id,name"},
                headers={"Accept": MediaType.JSON},
            )
            configurations = response["data"]
            for configuration in configurations:
                print(f'{configuration["id"]}: {configuration["name"]}')
            client.logout()

        # Retrieve configurations. Request that the response is in 'CSV' format.
        # The result should contain only the first 10, ordered alphabetically by name.
        with Client(<bam_host_url>) as client:
            client.login(<username>, <password>)
            configurations_csv = client.http_get(
                "/configurations",
                params={"orderBy": "asc(name)", "limit": "10"},
                headers={"Accept": MediaType.CSV},
            )
            configurations = list(csv.reader(configurations_csv.splitlines()))
            for configuration in configurations:
                # NOTE: The 'id' is the first value in a row, the 'name' is the third one.
                print(f"{configuration[0]}: {configuration[2]}")
            client.logout()

    .. versionadded:: 23.1.0
    """

    def __init__(self, url: str, *, verify: Any = True) -> None:
        u = parse_url(url)
        # if not u.scheme:
        #     raise ClientError("The BAM URL does not contain a scheme. It is required.")
        if not u.host:
            raise ClientError("The BAM URL does not contain a host. It is required.")

        super().__init__(_Instance(url))
        self.session.verify = verify
        self.__session_data = None
        self.__version = None  # The version of BAM.

    # region Inherited interfaces

    def _url_for(self, name: str) -> str:
        # NOTE: The `self.api_url_base` does not end with `/`.
        # The value of `name` is expected to start with `/`.
        return self.target.api_url_base.url + name

    def _handle_request_exc(self, exc: Exception) -> None:
        """
        Handle the exception that is raised when the request is attempted.

        :param exc: The exception raised when the HTTP request is made.
        :raises GeneralError: When the request to BlueCat Address Manager fails.

        This method is not intended to be called by users of the class. It is
        present to modify the behaviour inherited from the base class.
        """
        raise GeneralError("The request to BlueCat Address Manager REST API v2 failed.") from exc

    def _handle_error_response(self, response: requests.Response) -> None:
        """
        Handle a response that's considered to be an error (based on the HTTP
        status code). The standard implementation raises an exception.

        :param response: Object representing the HTTP response.

        .. versionchanged:: 23.3.0
        """
        response_content_type = response.headers.get("Content-Type")
        if response_content_type == "application/json":
            try:
                content = response.json()
            except Exception as exc:
                raise UnexpectedResponse(
                    "The BAM REST API v2 error response is not valid JSON,"
                    " despite the specified content type.",
                    response,
                ) from exc
            raise BAMV2ErrorResponse(
                message=content["message"],
                response=response,
                status=content["status"],
                code=content["code"],
                reason=content["reason"],
            )
        raise UnexpectedResponse(
            "The BAM REST API v2 error response does not have JSON content"
            f" type as expected: {response_content_type}",
            response,
        )

    def _handle_nonerror_response(self, response: requests.Response) -> Union[str, dict]:
        """
        Handle a response that's considered to be a success (non error), based
        on the HTTP status code. Depending on the content type of the response,
        its body is either parsed as JSON and the resulting object returned, or
        returned as verbatim text.

        :param response: Object representing the HTTP response.
        :return: The result of the handled response.
        """
        response_content_type = response.headers.get("Content-Type")
        content = (
            response.json()
            if response_content_type and "json" in response_content_type
            else response.text
        )
        return content

    def http_request(
        self,
        method: str,
        url: str,
        params: Optional[Union[dict, list[tuple], bytes]] = None,
        data: Optional[Union[dict, list[tuple], bytes, IO]] = None,
        headers: Optional[dict] = None,
        cookies: Optional[dict] = None,
        files: Optional[dict[str, IO]] = None,
        auth: Optional[Union[tuple, Callable]] = None,
        timeout: Optional[Union[float, tuple]] = None,
        allow_redirects: bool = True,
        proxies: Optional[dict] = None,
        hooks: Optional[dict] = None,
        stream: Optional[bool] = None,
        verify: Union[bool, str] = None,
        cert: Optional[Union[str, tuple[str, str]]] = None,
        json: Optional[Any] = None,
        expected_status_codes: Optional[Iterable[int]] = None,
    ) -> Union[dict, str]:
        """
        Perform an HTTP request to the Address Manager using the provided
        parameters.

        The call will be made through the instance of ``requests.Session`` that
        the client maintains. The parameters have the same behaviour as their
        namesakes in the ``requests`` library. The only exception is ``url``,
        whose value should be relative to the Address Manager URL, e.g.,
        ``"/configurations"``.

        :param method: HTTP method to be used, e.g., ``GET``, ``POST``,
            ``PUT``, ``PATCH``, ``DELETE``.
        :param url: A URL, relative to the API root to be used for the request,
            e.g., ``/configurations``. It must start with a forward slash.
        :param params: Query parameters to be passed in the request.
        :param data: Value to send in the body of the request. It can be
            a dictionary, list of tuples, bytes, or file-like object.
        :param headers: HTTP headers to send with the request.
        :param cookies: Cookies to send with the request.
        :param files: File-like objects for multipart encoding upload, e.g.
            ``{"filename": file_like_object}``
        :param auth: Object to handle HTTP Authentication.
        :param timeout: How long to wait to receive data before giving up. If
            given as a tuple, it is used as (connect, read) timeouts.
        :param allow_redirects: Whether to allow redirects. Defaults to
            ``True``.
        :param proxies: Mapping of protocol or protocol and hostname to a URL
            of a proxy to be used.
        :param hooks: Hooks to be called upon events.
        :param stream: Whether to immediately read the response content.
            Defaults to ``False``.
        :param verify: Whether to verify the server's TLS certificate. If a
            string is passed, it is treated as a path to a CA bundle to be
            used. Defaults to ``True``.
        :param cert: Path to a SSL client certificate file (.pem). If the
            certificate has a key, it can be provides as the second item in a
            tuple, with the first item being the path to the certificate file.
        :param json: Object to be sent as JSON in the body of the request.
        :param expected_status_codes: HTTP status codes that are acceptable as
            a status code of the HTTP response. If the received code does not
            match the passed values, an :py:exc:`UnexpectedResponse` is raised.
            If left empty, the status code validation is not performed.
        :return: The data the API responded with. The client will process the
            response based on the stated content type and return either the
            JSON formatted data parsed into a Python object (e.g., a ``dict``)
            or the verbatim body text.

        Example:

        .. code-block:: python

            from bluecat_libraries.address_manager.apiv2 import Client

            # Retrieve the configurations. Request the data as per BAM's default content type.
            with Client(<bam_host_url>) as client:
                client.login(<username>, <password>)
                response = client.http_request("GET", "/configurations")
                configurations = response["data"]
                for configuration in configurations:
                    print(f'{configuration["id"]}: {configuration["name"]}')
                client.logout()
        """
        url = self._url_for(url)
        return super().http_request(
            method,
            url=url,
            params=params,
            data=data,
            headers=headers,
            cookies=cookies,
            files=files,
            auth=auth,
            timeout=timeout,
            allow_redirects=allow_redirects,
            proxies=proxies,
            hooks=hooks,
            stream=stream,
            verify=verify,
            cert=cert,
            json=json,
            expected_status_codes=expected_status_codes,
        )

    def http_get(
        self,
        url: str = "",
        params=None,
        expected_status_codes: Optional[Iterable[int]] = None,
        **kwargs,
    ) -> Union[dict, str]:
        """
        Perform an HTTP ``GET`` request based on the provided parameters.
        See method ``http_request`` for details.

        If ``expected_status_codes`` is not specified, it defaults to ``(200, )``.
        ``200`` is the HTTP status code for successful response ``OK``.

        .. note:: From BAM v9.5.0, we can use the GET request to filter data
            without ``url`` in the prefix.
        """
        if expected_status_codes is None:
            expected_status_codes = (requests.codes.ok,)
        return self.http_request(
            "GET", url, params, expected_status_codes=expected_status_codes, **kwargs
        )

    def http_post(
        self,
        url: str,
        params=None,
        data=None,
        json: Optional[Any] = None,
        expected_status_codes: Optional[Iterable[int]] = None,
        **kwargs,
    ) -> Union[dict, str]:
        """
        Perform an HTTP ``POST`` request based on the provided parameters.
        See method ``http_request`` for details.
        """
        return self.http_request(
            "POST",
            url,
            params,
            data,
            json=json,
            expected_status_codes=expected_status_codes,
            **kwargs,
        )

    def http_put(
        self,
        url: str,
        params=None,
        data=None,
        json: Optional[Any] = None,
        expected_status_codes: Optional[Iterable[int]] = None,
        **kwargs,
    ) -> Union[dict, str]:
        """
        Perform an HTTP ``PUT`` request based on the provided parameters.
        See method ``http_request`` for details.
        """
        return self.http_request(
            "PUT",
            url,
            params,
            data,
            json=json,
            expected_status_codes=expected_status_codes,
            **kwargs,
        )

    def http_patch(
        self,
        url: str,
        params=None,
        data=None,
        json: Optional[Any] = None,
        expected_status_codes: Optional[Iterable[int]] = None,
        **kwargs,
    ) -> Union[dict, str]:
        """
        Perform an HTTP ``PATCH`` request based on the provided parameters.
        See method ``http_request`` for details.
        """
        return self.http_request(
            "PATCH",
            url,
            params,
            data,
            json=json,
            expected_status_codes=expected_status_codes,
            **kwargs,
        )

    def http_delete(
        self,
        url: str,
        params=None,
        data=None,
        json: Optional[Any] = None,
        expected_status_codes: Optional[Iterable[int]] = None,
        **kwargs,
    ) -> Union[dict, str]:
        """
        Perform an HTTP ``DELETE`` request based on the provided parameters.
        See method ``http_request`` for details.
        """
        return self.http_request(
            "DELETE",
            url,
            params,
            data,
            json=json,
            expected_status_codes=expected_status_codes,
            **kwargs,
        )

    # endregion Inherited interfaces
    # region BAM REST v2 API specific interfaces

    @property
    def auth(self) -> str | None:
        """
        The value that the client is configured to use for header ``Authorization`` when making
        requests to BAM REST v2 API.
        """
        return self.session.headers.get("Authorization", None)

    @auth.setter
    def auth(self, value: str) -> None:
        """
        Set the value to use for header ``Authorization``. It is used for authenticating the
        requests to the BAM REST v2 API.

        :param value: Token to be used for BAM authentication.
        """
        # Added in the initial version: 23.1.0
        if not value:
            raise ClientError(
                "Empty value for authentication to BAM REST v2 API."
                " If you need to clear the authentication, use the deleter."
            )

        self.__session_data = self.http_get("/sessions/current", headers={"Authorization": value})
        pos = value.find(" ")
        if pos > -1:
            self.token_type = value[:pos]
            self.token = value[pos + 1 :]
        else:
            self.token_type = None
            self.token = value
        self.session.headers.update({"Authorization": value})

    @auth.deleter
    def auth(self) -> None:
        """Clear the value to use for header ``Authorization``."""
        # Added in the initial version: 23.1.0
        self.token_type = None
        self.token = None
        self.session.headers.pop("Authorization", None)

    @property
    def bam_url(self) -> str:
        """
        A URL to the Address Manager the client is created for.

        This is not the verbatim value given during initialization. It contains
        only the scheme, host, and (if specified) port, which will be used by
        the client.
        """
        # Added in the initial version: 23.1.0
        return self.target.url.url

    @property
    def bam_version(self) -> version.Version:
        """
        The version of the BlueCat Address Manager the client is connected to.

        .. note:: The client has to be authenticated before the version can be
            determined.

        :return: Version of BlueCat Address Manager.
        """
        # Added in the initial version: 23.1.0
        if not self.__version:
            params = {"filter": "type:'SystemSettings'"}
            response = self.http_get("/settings", params=params)
            try:
                v = response["data"][0].get("version").split("-")[0]
            except Exception as exc:
                raise UnexpectedResponse(
                    "The BAM response does not contain a version value formatted as expected.",
                    self._last_response,
                ) from exc
            self.__version = version.parse_version(v)
        return self.__version

    @property
    def is_authenticated(self) -> bool:
        """
        Whether the authentication necessary to communicate with the BlueCat
        Address Manager REST v2 API is set.
        """
        # Added in the initial version: 23.1.0
        return bool(self.session.headers.get("Authorization", False))

    def login(self, username: str, password: str) -> dict:
        """
        Log user into BAM and store authentication token for later use by subsequent calls.

        :param username: BAM username.
        :param password: BAM password.

        :return: The session information returned by BAM API v2 basic authentication token.
        """
        # Added in the initial version: 23.1.0
        data = {"username": username, "password": password}
        self.__session_data = self.http_post("/sessions", json=data)
        self.auth = "Basic " + self.__session_data["basicAuthenticationCredentials"]
        return copy.deepcopy(self.__session_data)

    def logout(self) -> dict:
        """
        Log user out from BAM and clear any stored authentication token.

        :return: The logged-out session information is returned by BAM API v2.
        """
        # Added in the initial version: 23.1.0
        sessions_id = self.__session_data["id"]
        data = {"state": "LOGGED_OUT"}
        headers = {"Content-Type": MediaType.MERGE_PATCH_JSON}
        res = self.http_patch(f"/sessions/{sessions_id}", json=data, headers=headers)
        self.__session_data = None
        del self.auth
        return res

    # endregion BAM REST v2 API specific actions
